# -*- encoding: utf8 -*-
import datetime
import random
import shelve
import pickle

from django.shortcuts import render_to_response
from django.http import HttpResponseRedirect, HttpResponse
from django.core.urlresolvers import reverse
from django.conf import settings
from django.template import Template
from django.template.context import Context

from dihabs.gtools import build_groups, gset
from dihabs.util import adify, impmod
from dihabs.settings import RESULTS, TESTMODE, SESSION_DEBUG, INTERVIEW
from dihabs.templatetags.xfilters import format

interview = impmod(INTERVIEW)
iva = adify(interview.spec) # attribute style access to interview
messages = adify(iva.messages)

RQS_RESULT_ITEMS = ('start', 'disclosures', 'npersons', 'nvalues', 'groups',
                    'epilog', 'bsid')

# =============================================================================
# utilities
# =============================================================================

# shortcuts to write/read/delete python objects to/from cookies
wcookie = lambda rp, k, o: rp.set_cookie(k, pickle.dumps(o, protocol=0)) # as long as session cookie
rcookie = lambda rq, k: pickle.loads(rq.COOKIES[k])
dcookie = lambda rp, k: rp.delete_cookie(k)

def _rcurrent(rqs):
    """HTTP redirect response to current step."""

    url = reverse('dihabs.views.cstep', args=(rqs['step'],))
    return HttpResponseRedirect(url)

def _irqc(request):
    """Initialize a request context."""

    rqc = {}
    rqc['session'] = request.session
    rqc['MEDIA_URL'] = settings.MEDIA_URL
    rqc['msg'] = messages
    rqc['step'] = request.session.get('step', -2)

    return rqc

# =============================================================================
# views
# =============================================================================

def entry(request):
    """Inits session and immediatelty redirects to cookie check."""

    rqs = request.session
    rqs.set_expiry(0)

    # if there is a session, redirect to current step
    if 'usid' in rqs:
        return _rcurrent(rqs)

    # otherwise create a new session and redirect to cookie-check view
    rqs['usid'] = "%032x" % random.randint(0, 2**128)
    rqs['ngroup'] = 0 # number of current group
    rqs['nitype'] = 0 # number of current itype
    ngroups = isinstance(iva.groups[0], int) and sum(iva.groups) or len(iva.groups)
    stepmap, i = {}, 0
    for i in range(i+1, i+1 + 1): stepmap[i] = _start
    for i in range(i+1, i+1 + 1): stepmap[i] = _setnames
    for i in range(i+1, i+1 + len(iva.itypes)): stepmap[i] = _setvalues
    for i in range(i+1, i+1 + ngroups + iva.ngdups): stepmap[i] = _disclose
    for i in range(i+1, i+1 + 1): stepmap[i] = _epilog
    for i in range(i+1, i+1 + 1): stepmap[i] = _finish
    rqs['stepmap'] = stepmap
    rqs['steps'] = len(stepmap)
    rqs['step'] = 1
    rqs['disclosures'] = [] # interview disclosures
    rqs['start'] = datetime.datetime.now() # start of interview
    rqs['debug'] = SESSION_DEBUG
    rqs['npersons'] = 0 # number of persons given by interviewee
    rqs['nvalues'] = [] # list of number of values per itype
    rqs['bsid'] = rqs.session_key
    r = HttpResponseRedirect(reverse('dihabs.views.cookiecheck'))
    wcookie(r, 'names', [])
    wcookie(r, 'values', {})
    return r

def cookiecheck(request):
    """Checks if cookies are available and then redirects to step dispatcher."""

    rqs = request.session

    # if cookies work (session is available), redirect to survey view
    if 'usid' in rqs:
        return _rcurrent(rqs)

    # otherwise show a cookie error message
    return render_to_response('dihabs/nocookie.html')

def lost(request):
    """Current step dispatcher"""

    rqs = request.session
    if 'usid' not in rqs:
        return HttpResponseRedirect(reverse('dihabs.views.entry'))
    return _rcurrent(rqs)

def cstep(request, step):
    """Current step dispatcher"""

    rqs = request.session

    if 'usid' not in rqs:
        return HttpResponseRedirect(reverse('dihabs.views.entry'))
    step = int(step)
    if step != rqs['step']:
        if request.method == "POST":
            rqc = _irqc(request)
            rqc['html'] = format(messages.badstep)
            rqc['action'] = reverse('dihabs.views.cstep', args=(rqs['step'],))
            rqc['method'] = "GET"
            rqc['submit'] = messages.badstep_ack
            return render_to_response('dihabs/simple.html', rqc)
        return _rcurrent(rqs)

    # At least one session field must be assigned, otherwise it does not get
    # saved and changes within already assigned objects get lost !!!
    rqs['dummy'] = datetime.datetime.now()

    handler = rqs['stepmap'][step]

    return handler(request)

def reset(request):

    request.session.clear()

    return HttpResponseRedirect(reverse('dihabs.views.entry'))

# =============================================================================
# step handler
# =============================================================================

def _start(request):
    """Welcome message."""

    rqs = request.session
    rqc = _irqc(request)

    # get needed state data and redirect to entry if it is not available

    if request.method == "POST":
        rqs['step'] += 1
        return _rcurrent(rqs)

    rqc['html'] = format(messages.start)
    rqc['action'] = reverse('dihabs.views.cstep', args=(rqs['step'],))
    rqc['method'] = "POST"
    rqc['submit'] = messages.start_ack
    return render_to_response('dihabs/simple.html', rqc)

def _setnames(request):
    """Set person names."""

    rqs = request.session
    rqc = _irqc(request)

    if request.method == 'POST':
        names = set()
        for k,v in request.POST.items():
            v = v.strip()
            if k.startswith('name_') and v:
                names.add(v)
        names = sorted(names)
        if len(names) < iva.npersons_min:
            rqs['xerror'] = messages.setnames_need_more % iva.npersons_min
            rqs['xnames'] = names + ([''] * (iva.npersons_max - len(names)))
            return _rcurrent(rqs)
        if len(names) > iva.npersons_max_sample:
            names = random.sample(names, iva.npersons_max_sample)
        random.shuffle(names)
        rqs['npersons'] = len(names)
        rqs['groups'] = build_groups(len(names), iva.groups)
        duplicates = random.sample(rqs['groups'], iva.ngdups)
        random.shuffle(rqs['groups'])
        rqs['groups'] += duplicates
        rqs['step'] += 1
        r = _rcurrent(rqs)
        wcookie(r, 'names', names)
        return r

    else: ### GET request ###

        rqc['progress'] = int(float(rqs['step']) / rqs['steps'] * 100)
        rqc['names'] = rqs.pop('xnames', [''] * iva.npersons_max)
        rqc['error'] = format(rqs.pop('xerror', ""))
        return render_to_response('dihabs/setnames.html', rqc)

def _setvalues(request):
    """Set items for a specific information type."""

    rqs = request.session
    rqc = _irqc(request)

    itype = iva.itypes[rqs['nitype']]

    if request.method == 'POST':

        values = set()
        for k,v in request.POST.items():
            v = v.strip()
            if k.startswith('value_') and v:
                values.add(v)
        values = sorted(values)

        if len(values) < itype.nvalues_min:
            rqs['xerror'] = messages.setvalues_need_more % itype.nvalues_min
            rqs['xvalues'] = values + ([""] * (itype.nvalues_max - len(values)))
            return _rcurrent(rqs)

        rqs['nvalues'].append(len(values))
        rqs['nitype'] += 1
        rqs['step'] += 1

        r = _rcurrent(rqs)

        vcookie = rcookie(request, 'values')
        vcookie[itype.name] = values
        wcookie(r, 'values', vcookie)
        return r

    else: ### GET request ###

        rqc['progress'] = int(float(rqs['step']) / rqs['steps'] * 100)
        rqc['name'] = itype.name
        rqc['step'] = rqs['step']
        rqc['desc'] = format(itype.desc_input % itype.nvalues_min)
        rqc['values'] = rqs.pop("xvalues", [""] * itype.nvalues_max)
        rqc['error'] = format(rqs.pop("xerror", ""))
        return render_to_response('dihabs/setvalues.html', rqc)

def _disclose(request):
    """Disclose information items to a specific person group."""

    rqs = request.session
    rqc = _irqc(request)

    ### process form data ###

    if request.method == 'POST':

        disclosure = {}
        disclosure['start'] = rqs['disclosure-start']
        disclosure['end'] = datetime.datetime.now()
        disclosure['group'] = rqs['groups'][rqs['ngroup']]
        selected = [tuple(int(i) for i in k.split('_')[1:])
                    for k in request.POST if k.startswith('value_')]
        disclosure['values'] = 0
        disclosure['values-per-type'] = [0] * len(rqs['nvalues'])
        for i, j in selected:
            disclosure['values'] += 2**(sum(rqs['nvalues'][:i]) + j)
            disclosure['values-per-type'][i] += 2**j
        rqs['disclosures'].append(disclosure)
        rqs['ngroup'] += 1
        rqs['step'] += 1
        return _rcurrent(rqs)

    ### setup request context and show current disclosure question ###

    rqs['disclosure-start'] = datetime.datetime.now()
    rqc['progress'] = int(float(rqs['step']) / rqs['steps'] * 100)
    group = gset(rqs['groups'][rqs['ngroup']])
    names = rcookie(request, 'names')
    rqc['names'] = [name for n, name in enumerate(names) if n in group]
    values = rcookie(request, 'values')
    rqc['itypes'] = []
    for itype in iva.itypes:
        rqc['itypes'].append((itype.name, format(itype.desc_select), values[itype.name]))

    return render_to_response('dihabs/disclose.html', rqc)

def _epilog(request):
    """Epilog, as set in interview spec."""

    rqs = request.session
    rqc = _irqc(request)

    ### process form data ###

    if request.method == 'POST':
        errors, data = interview.process_epilog_formdata(request.POST)
        if errors:
            # TODO
            pass
        rqs['epilog'] = data
        rqs['step'] += 1
        return _rcurrent(rqs)

    else: ### GET request ###

        tmpl = interview.get_epilog_template(rqc)
        if not tmpl:
            rqs['step'] += 1
            return _rcurrent(rqs)
        rqc['progress'] = int(float(rqs['step']) / rqs['steps'] * 100)
        rqc['error'] = format(rqs.pop('xerror', ""))
        content = Template(tmpl).render(Context(rqc))
        return HttpResponse(content)
        #return render_to_response(interview.epilog_template, rqc)

def _finish(request):
    """Save and reset session data."""

    rqs = request.session
    rqc = _irqc(request)
    rqc['session'] = dict(rqs) # rqs is cleared below -> we need a copy here

    result = {}
    result['end'] = datetime.datetime.now()
    for key in RQS_RESULT_ITEMS:
        result[key] = rqs[key]
    result['interview'] = interview.spec
    result['test'] = TESTMODE

    results = shelve.open(RESULTS, protocol=2)
    results[rqs['usid']] = result
    results.close()

    rqc['progress'] = int(float(rqs['step']) / rqs['steps'] * 100)

    rqs.clear()

    tmpl = interview.get_finish_template(rqc, result)
    r = HttpResponse(Template(tmpl).render(Context(rqc)))
    wcookie(r, 'names', [])  # deleting would be better, but entry view does
    wcookie(r, 'values', {}) # not always reset them properly

    return r
